; Author: Petroizki
	
; NOTE! To save some calculations i usually compare the distance on square,
;  this means that i don't take the square root from the distance function (sqrt(x*X + y*y)).
	
	AMOUNT_OF_DOTS EQU <666>		;! must be >0 !!!
	DISTANCE_CURSOR_SQRT EQU <2500.0>; minimum distance to cursor (pow2)
	DISTANCE_MINIMUM EQU <30.0>		; minimum distance to other dots
	DISTANCE_NEIGHBOUR_SQRT EQU <7000.0> ; distance within dots are threathed as neighbours (pow2)
	SPEED_EXTRA_RECIPROCAL EQU <0.7>; extra speed that is added, when cursor is near dot
	SPEED_NORMAL_RECIPROCAL EQU <0.4>; normal speed (1/true_speed)
	WALL_DISTANCE EQU <40.0>			; distance to walls the dots try to keep
	WALL_MIN_DISTANCE EQU <15>		; absolute minimum distance to the walls
	WEIGHT_CURSOR EQU <1.3>			; "weight" (strength) of cursor vector, which makes dots avoid cursor

	MATRIX2x1 struct
		x REAL4 ?
		y REAL4 ?
	MATRIX2x1 ends
	
	SINGLE_DOT struct
		color COLORREF ?			; color of the dot
		plan MATRIX2x1 <?>			; future plans
		position MATRIX2x1 <?>		; position of the dot
		velocity MATRIX2x1 <?>		; velocity of the dot
	SINGLE_DOT ends
	
	CreateDots proto :DWORD
	RunAI proto :REAL4, :REAL4
	UpdateWorld proto :HDC
.data
	fDistanceCursor REAL4 DISTANCE_CURSOR_SQRT
	fDistanceMinimum REAL4 DISTANCE_MINIMUM
	fDistanceNeighbour REAL4 DISTANCE_NEIGHBOUR_SQRT
	fDistanceWalls REAL4 WALL_DISTANCE
	fSpeedExtra REAL4 SPEED_EXTRA_RECIPROCAL
	fSpeedNormal REAL4 SPEED_NORMAL_RECIPROCAL
	fWeightCursor REAL4 WEIGHT_CURSOR
.data?
	fMaxX REAL4 ?
	fMaxY REAL4 ?
	dwMaxX DWORD ?
	dwMaxY DWORD ?
	dots SINGLE_DOT AMOUNT_OF_DOTS dup (<?>)
.code
; creates dots with random position and color
CreateDots proc uses ebx edi, WorldRect:DWORD
	LOCAL temp:DWORD
	mov edx, WorldRect							; get world rect
	
	mov eax, dword ptr [edx].RECT.right		; get right wall pos
	sub eax, WALL_MIN_DISTANCE					; subtract the minimum distance to wall
	mov dword ptr [dwMaxX], eax			 	; save the maximum x pos
	
	mov eax, dword ptr [edx].RECT.bottom		; get bottom wall pos
	sub eax, WALL_MIN_DISTANCE					; subtract the minimum distance to wall
	mov dword ptr [dwMaxY], eax				; save the maximum y pos
	
	; here we make the "second" wall minimum distances
	fild dword ptr [edx].RECT.right
	fsub dword ptr [fDistanceWalls]
	fstp dword ptr [fMaxX]
	
	fild dword ptr [edx].RECT.bottom
	fsub dword ptr [fDistanceWalls]
	fstp dword ptr [fMaxY]
	
	mov edi, OFFSET dots
	mov ebx, AMOUNT_OF_DOTS
	
	; creates random colored dot to random position
	.REPEAT
		mov dword ptr [edi].SINGLE_DOT.velocity.x, 40000000h
		mov dword ptr [edi].SINGLE_DOT.velocity.y, 40000000h
		mov dword ptr [edi].SINGLE_DOT.plan.x, 40000000h
		mov dword ptr [edi].SINGLE_DOT.plan.y, 40000000h
		
		invoke nrandom, 0FFFFFFh
		mov dword ptr [edi].SINGLE_DOT.color, eax
		
		invoke nrandom, SCREEN_WIDTH
		
		push eax
		fild dword ptr [esp]
		fstp dword ptr [edi].SINGLE_DOT.position.x
		mov dword ptr [esp], SCREEN_HEIGHT
		call nrandom
		
		mov temp, eax
		fild dword ptr [temp]
		fstp dword ptr [edi].SINGLE_DOT.position.y
		
		add edi, SIZEOF SINGLE_DOT
		dec ebx
	.UNTIL zero?
	
	ret
CreateDots endp

; runs the A.I. for all dots, this will set every dots velocity
RunAI proc uses ebx edi esi, fMousePosX:REAL4, fMousePosY:REAL4
	LOCAL temp:DWORD
	; forces that affect the dots:
	; 1. Alignment
	;     Dot will align the velocity vector with local flock (neighborhood dots).
	; 2. Separation
	;     Dot will keep a minimum distance from local dots.
	; 3. Cohesion
	;     Dot will try to fly the center position of local flock.
	; 4. Evasion
	;     Dot will try to evade mouse cursor.
	;     Dot will evade world edges.
	;      there are two world edge checks, other one creates an vector
	;      from the wall, and hints the dot to go another way,
	;      the second check just forces the dot to stay within the window
	
	mov ecx, AMOUNT_OF_DOTS
	mov edx, OFFSET dots
	
	.REPEAT
		xor ebx, ebx
		
		; these will be here troughout the inner loop
		; the wall edge checker will pop them from the stack
		fld dword ptr [edx].SINGLE_DOT.position.x
		fld dword ptr [edx].SINGLE_DOT.position.y
		
		lea edi, [ecx - 1]
		
		test edi, edi
		jz @end_of_neighbours
		
		lea esi, [edx + SIZEOF SINGLE_DOT]
		
		fld dword ptr [edx].SINGLE_DOT.plan.x
		fld dword ptr [edx].SINGLE_DOT.plan.y
@neighbour_check:
			fld dword ptr [esi].SINGLE_DOT.position.x
			fsub st(0), st(4)
			
			fld dword ptr [esi].SINGLE_DOT.position.y
			fsub st(0), st(4)
			
			fld st(1)
			fmul st(0), st(0)
			fld st(1)
			fmul st(0), st(0)
			
			faddp st(1), st(0)
			
			fld dword ptr [fDistanceNeighbour]
			fcomip st(0), st(1)
			fwait
			jb @not_neighbour
			
			fldz
			fcomip st(0), st(1)
			fwait
			je @not_neighbour
			
			fsqrt
			fdiv st(1), st(0)
			fdiv st(2), st(0)
			
			; The dot is aligned the same way as dots nearby
			fld dword ptr [esi].SINGLE_DOT.plan.x
			fsub st(0), st(3)
			fadd dword ptr [edx].SINGLE_DOT.velocity.x
			fstp dword ptr [esi].SINGLE_DOT.plan.x
			
			fld dword ptr [esi].SINGLE_DOT.velocity.x
			faddp st(5), st(0)
			
			fld dword ptr [esi].SINGLE_DOT.plan.y
			fsub st(0), st(2)
			fadd dword ptr [edx].SINGLE_DOT.velocity.y
			fstp dword ptr [esi].SINGLE_DOT.plan.y
			
			fld dword ptr [esi].SINGLE_DOT.velocity.y
			faddp st(4), st(0)
			
			; dot will either wants to either pursue or avoid, the dot
			; nearby, depending how close it is
			; If the other dot is too close, we must go away
			fld dword ptr [fDistanceMinimum]
			fcomip st(0), st(1)
			fstp st(0)
			fwait
			jb @not_near
			
			fsubp st(2), st(0)
			fsubp st(2), st(0)
			jmp @F
@not_near:
			faddp st(2), st(0)
			faddp st(2), st(0)
			jmp @F
@not_neighbour:
			fstp st(0)
			fcompp
@@:
			add esi, SIZEOF SINGLE_DOT
			dec edi
		jnz @neighbour_check
		
		; The plan is normalized here
		fld st(1)
		fmul st(0), st(0)
		fld st(1)
		fmul st(0), st(0)
		faddp st(1), st(0)
		fsqrt
		fdiv st(2), st(0)
		fdivp st(1), st(0)
@end_of_neighbours:
;		int 3h

		; - Dot will try to evade walls
		;    Here we create an unit vector from wall, which is
		;    too near, and add it to the plan
		fld dword ptr [fDistanceWalls]
		fcomi st(0), st(4)
		fwait
		jb @check_right_wall
			fld1
			faddp st(2), st(0)
			jmp @check_bottom_wall
@check_right_wall:
		fld dword ptr [fMaxX]
		fcomip st(0), st(5)
		fwait
		ja @check_bottom_wall
			fld1
			fsubp st(3), st(0)
@check_bottom_wall:
		fld dword ptr [fMaxY]
		fcomip st(0), st(4)
		fwait
		ja @check_top_wall
			fld1
			fsubp st(2), st(0)
			fstp st(0)
			jmp @F
@check_top_wall:
		fcomip st(0), st(3)
		fwait
		jb @F
			fld1
			faddp st(1), st(0)
@@:
		fstp dword ptr [edx].SINGLE_DOT.plan.y
		fstp dword ptr [edx].SINGLE_DOT.plan.x
		
		; - Dot will evade mouse cursor
		;    we will just create vector to opposite direction to cursor,
		;    and add it to the final plan
		fld st(1)
		fsub dword ptr [fMousePosX]
		fld st(1)
		fsub dword ptr [fMousePosY]
		fld st(1)
		fmul st(0), st(0)
		fld st(1)
		fmul st(0), st(0)
		faddp st(1), st(0)
		
		fld dword ptr [fDistanceCursor]
		fcomip st(0), st(1)
		fwait
		jb @no_mouse_cursor
		
		; if mouse cursor is exactly in the same position,
		; then it has no affect
		fldz
		fcomip st(0), st(1)
		fwait
		je @no_mouse_cursor
		
		fsqrt
		fdiv st(2), st(0)
		fdivp st(1), st(0)
		
		fmul dword ptr [fWeightCursor]				; add the "weight" of mouse cursor
		fadd dword ptr [edx].SINGLE_DOT.plan.y
		fstp dword ptr [edx].SINGLE_DOT.plan.y
		
		fmul dword ptr [fWeightCursor]				; add the "weight" of mouse cursor
		fadd dword ptr [edx].SINGLE_DOT.plan.x
		fstp dword ptr [edx].SINGLE_DOT.plan.x
		or ebx, -1
		jmp @F
@no_mouse_cursor:
		fcompp
		fstp st(0)
@@:	
		; - Dot will evade world edges.
		;    Below is the last way to handle world walls
		;    The velocity vector element is just pointed to another direction,
		;    if dot is going over the world edges.
		fistp dword ptr [temp]					; save the y pos
		fwait
		mov eax, dword ptr [temp]
		
		.IF eax < WALL_MIN_DISTANCE
			and dword ptr [edx].SINGLE_DOT.plan.y, 7FFFFFFFh
		.ELSEIF eax > dword ptr [dwMaxY]
			or dword ptr [edx].SINGLE_DOT.plan.y, 80000000h
		.ENDIF
		
		fistp dword ptr [temp]					; save the x pos
		fwait
		mov eax, dword ptr [temp]
		
		.IF eax < WALL_MIN_DISTANCE
			and dword ptr [edx].SINGLE_DOT.plan.x, 7FFFFFFFh
		.ELSEIF eax > dword ptr [dwMaxX]
			or dword ptr [edx].SINGLE_DOT.plan.x, 80000000h
		.ENDIF
		
		; - The plan will be made as the current velocity,
		;    after it has been normalized and scaled to desired speed
		fld dword ptr [edx].SINGLE_DOT.plan.x
		fadd dword ptr [edx].SINGLE_DOT.velocity.x
		fld dword ptr [edx].SINGLE_DOT.plan.y
		fadd dword ptr [edx].SINGLE_DOT.velocity.y
		
		fld st(1)
		fmul st(0), st(0)
		fld st(1)
		fmul st(0), st(0)
		faddp st(1), st(0)
		fsqrt
		
		test ebx, ebx								; check if cursor is nearby
		jz @F
		fmul dword ptr [fSpeedExtra]				; add extra speed
@@:		fmul dword ptr [fSpeedNormal]				; scale vector with normal speed
		fdiv st(2), st(0)
		fdivp st(1), st(0)
		
		; store the new velocity of dot
		fstp dword ptr [edx].SINGLE_DOT.velocity.y
		fstp dword ptr [edx].SINGLE_DOT.velocity.x
		
		; empty the plan vector
		mov [edx].SINGLE_DOT.plan.x, 0
		mov [edx].SINGLE_DOT.plan.y, 0
		
		add edx, SIZEOF SINGLE_DOT
		dec ecx
	.UNTIL zero?
	
	ret
RunAI endp

; updates dots position, and draws them
UpdateWorld proc uses ebx edi esi, hDC:HDC
	mov ebx, AMOUNT_OF_DOTS
	mov esi, hDC
	mov edi, OFFSET dots
	
	.REPEAT
		push dword ptr [edi.SINGLE_DOT.color]		; 3
		push ecx									; 2
		
		fld dword ptr [edi.SINGLE_DOT.position.y]	; get pos x
		fist dword ptr [esp]						; 2 - store pos.x as integer
		fadd dword ptr [edi.SINGLE_DOT.velocity.y]	; update x pos
		fstp dword ptr [edi.SINGLE_DOT.position.y]	; store updated position
		
		push ecx									; 1
		
		fld dword ptr [edi.SINGLE_DOT.position.x]	; get pos y
		fist dword ptr [esp]						; 1 - store pos.y as integer
		fadd dword ptr [edi.SINGLE_DOT.velocity.x]	; update y pos
		fstp dword ptr [edi.SINGLE_DOT.position.x]	; store updated position
		
		push esi									; 0 - DC handle
		call SetPixel								; draw dot
		
		add edi, SIZEOF SINGLE_DOT					; goto next dot struct
		dec ebx										; decrease counter
	.UNTIL zero?
	
	ret
UpdateWorld endp